# Standard library imports.
import sys

# Major package imports.
import wx

# Local imports.
from omnivore import get_image_path

import logging
log = logging.getLogger(__name__)


class MouseHandler(object):
    """
    Processing of mouse events, separate from the rendering window
    
    This is an object-based control system of mouse modes
    """
    icon = "help.png"
    menu_item_name = "Generic Mouse Handler"
    menu_item_tooltip = "Tooltip for generic mouse handler"
    editor_trait_for_enabled = ""

    mouse_too_close_pixel_tolerance = 5

    def __init__(self, window):
        self.control = window
        self.snapped_point = None, 0
        self.first_mouse_down_position = 0, 0
        self.after_first_mouse_up = False
        self.mouse_up_too_close = False
        self.can_snap = False

        # left_up events can be called after double click on some platforms, so
        # this parameter can be set by the parent of the handler if the handler
        # wants to differentiate response to left_up
        self.num_clicks = 0

        # Optional (only OS X at this point) mouse wheel event filter
        self.wheel_scroll_count = 0
        self.use_every_nth_wheel_scroll = 5

        self.init_post_hook()

    def init_post_hook(self):
        pass

    def get_cursor(self):
        return wx.Cursor(wx.CURSOR_ARROW)

    def process_left_down(self, evt):
        evt.Skip()

    def get_position(self, evt):
        return evt.GetPosition()

    def process_mouse_motion_up(self, evt):
        self.control.release_mouse()
        evt.Skip()

    def process_mouse_motion_down(self, evt):
        evt.Skip()

    def reset_early_mouse_params(self):
        self.mouse_up_too_close = False
        self.after_first_mouse_up = False

    def check_early_mouse_release(self, evt):
        c = self.control
        p = evt.GetPosition()
        dx = p[0] - self.first_mouse_down_position[0]
        dy = p[1] - self.first_mouse_down_position[1]
        tol = self.mouse_too_close_pixel_tolerance
        if abs(dx) < tol and abs(dy) < tol:
            return True
        return False

    def process_left_up(self, evt):
        c = self.control
        c.release_mouse()  # it's hard to know for sure when the mouse may be captured
        c.selection_box_is_being_defined = False

    def process_left_dclick(self, evt):
        evt.Skip()

    def process_popup(self, evt):
        actions = self.get_popup_actions(evt)
        if actions:
            self.control.editor.popup_context_menu_from_actions(self.control, actions)

    def get_popup_actions(self, evt):
        return None

    def process_mouse_wheel(self, evt):
        c = self.control
        e = c.editor
        rotation = evt.GetWheelRotation()
        delta = evt.GetWheelDelta()
        window = evt.GetEventObject()
        log.debug("on_mouse_wheel_scroll. rot=%s delta=%d win=%s" % (rotation, delta, window))
        if rotation == 0 or delta == 0:
            return

        if sys.platform == "darwin":
            # OS X mouse wheel handling is not the same as other platform.
            # The delta value is 10 while the other platforms are 120,
            # and the rotation amount varies, seemingly due to the speed at
            # which the wheel is rotated (or how fast the trackpad is swiped)
            # while other platforms are either 120 or -120.  When mouse wheel
            # handling is performed in the usual manner on OS X it produces a
            # strange back-and-forth zooming in/zooming out.  So, this extra
            # hack is needed to operate like the other platforms.

            # add extra to the rotation so the minimum amount is 1 or -1
            extra = delta if rotation > 0 else -delta
            amount = (rotation + extra) / delta
            self.wheel_scroll_count -= abs(amount)
            if self.wheel_scroll_count > 0:
                return
            self.wheel_scroll_count = self.use_every_nth_wheel_scroll
        else:
            amount = rotation / delta

        if evt.ControlDown():
            self.zoom_mouse_wheel(evt, amount)
        else:
            self.pan_mouse_wheel(evt, amount)

    def zoom_mouse_wheel(self, evt, amount):
        pass

    def pan_mouse_wheel(self, evt, amount):
        evt.Skip()

    def process_mouse_enter(self, evt):
        evt.Skip()

    def process_mouse_leave(self, evt):
        evt.Skip()

    def process_focus(self, evt):
        pass

    def process_focus_lost(self, evt):
        pass

    def process_key_char(self, evt):
        c = self.control
        if (evt.GetKeyCode() == wx.WXK_ESCAPE):
            self.esc_key_pressed()
        elif evt.GetKeyCode() == wx.WXK_DELETE:
            self.delete_key_pressed()
        elif evt.GetKeyCode() == wx.WXK_BACK:
            self.backspace_key_pressed()
        else:
            evt.Skip()

    def esc_key_pressed(self):
        self.control.project.clear_all_selections()

    def delete_key_pressed(self):
        pass

    def backspace_key_pressed(self):
        pass



class MouseControllerMixin(object):
    """Mixin for a control to adapt to the MouseHandler class
    
    Must specify the default pan mode
    """

    def __init__(self, pan_mode_class):
        p = get_image_path("icons/hand.ico")
        self.hand_cursor = wx.Cursor(p, wx.BITMAP_TYPE_ICO, 16, 16)
        p = get_image_path("icons/hand_closed.ico")
        self.hand_closed_cursor = wx.Cursor(p, wx.BITMAP_TYPE_ICO, 16, 16)
        self.forced_cursor = None
        self.set_mouse_mode(MouseHandler)  # dummy initial mouse handler
        self.default_pan_mode = pan_mode_class(self)
        self.batch = None

    def set_mouse_mode(self, handler):
        self.release_mouse()
        self.mouse_mode = handler(self)

    def set_cursor(self, mode=None):
        if (self.forced_cursor is not None):
            self.SetCursor(self.forced_cursor)
            #
            return

        if mode is None:
            mode = self.mouse_mode
        c = mode.get_cursor()
        self.SetCursor(c)

    def get_effective_tool_mode(self, event):
        middle_down = False
        alt_down = False
        if (event is not None):
            try:
                alt_down = event.AltDown()
                # print self.is_alt_key_down
            except:
                pass
            try:
                middle_down = event.MiddleIsDown()
            except:
                pass
        if alt_down or middle_down:
            mode = self.default_pan_mode
        else:
            mode = self.mouse_mode
        return mode

    def release_mouse(self):
        self.mouse_is_down = False
        self.selection_box_is_being_defined = False
        while self.HasCapture():
            self.ReleaseMouse()

    def on_left_down(self, evt):
        # self.SetFocus() # why would it not be focused?
        mode = self.get_effective_tool_mode(evt)
        self.forced_cursor = None
        self.selection_box_is_being_defined = False
        self.mouse_down_position = evt.GetPosition()
        self.mouse_move_position = self.mouse_down_position

        mode.process_left_down(evt)
        self.set_cursor(mode)
        evt.Skip()

    def on_motion(self, evt):
        mode = self.get_effective_tool_mode(evt)
        log.debug("on_motion: effective mode=%s" % mode)
        if evt.LeftIsDown():
            mode.process_mouse_motion_down(evt)
        else:
            mode.process_mouse_motion_up(evt)
        self.set_cursor(mode)
        evt.Skip()

    def on_left_up(self, evt):
        mode = self.get_effective_tool_mode(evt)
        self.forced_cursor = None
        mode.process_left_up(evt)
        self.set_cursor(mode)
        evt.Skip()

    def on_left_dclick(self, evt):
        # self.SetFocus() # why would it not be focused?
        mode = self.get_effective_tool_mode(evt)
        mode.process_left_dclick(evt)
        self.set_cursor(mode)
        evt.Skip()

    def on_popup(self, evt):
        mode = self.get_effective_tool_mode(evt)
        self.forced_cursor = None
        mode.process_popup(evt)
        self.set_cursor(mode)

    def on_mouse_wheel(self, evt):
        mode = self.get_effective_tool_mode(evt)
        mode.process_mouse_wheel(evt)
        self.set_cursor(mode)

    def on_mouse_enter(self, evt):
        self.set_cursor()
        evt.Skip()

    def on_mouse_leave(self, evt):
        self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
        self.mouse_mode.process_mouse_leave(evt)
        evt.Skip()

    def on_key_char(self, evt):
        mode = self.get_effective_tool_mode(evt)
        self.set_cursor(mode)

        mode.process_key_char(evt)
        evt.Skip()

    def on_focus(self, evt):
        mode = self.get_effective_tool_mode(evt)
        mode.process_focus(evt)

    def on_focus_lost(self, evt):
        mode = self.get_effective_tool_mode(evt)
        mode.process_focus_lost(evt)




#------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# Copyright (c) 2015-2017, Enthought, Inc.
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in enthought/LICENSE.txt and may be redistributed only
# under the conditions described in the aforementioned license.  The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
#
#------------------------------------------------------------------------------
"""
This module provides the toolkit object for the current backend toolkit

When first imported this module will use the 'pyface.toolkit' entry points
from pkg_resources to load a plugin according to the following strategy:

- if ETSConfig.toolkit is set, try to load a plugin with a matching name.
  If it succeeds, we are good, and if it fails then we error out.
- if ETSConfig.toolkit is not set, we try to load the 'qt4' and 'wx' plugins
  in that order, and on success we consider ourselves good.  The
  ETSConfig.toolkit gets set appropriately.
- after that, we try every 'pyface.toolkit' plugin we can find except 'null'
  in an arbitrary order.  If one succeeds, we consider ourselves good, and
  set the ETSConfig.toolkit appropriately.
- after that, try the 'null' plugin.
- finally, if all else fails, we try to directly import the null backend's
  toolkit object.
"""

import logging

import pkg_resources

from traits.etsconfig.api import ETSConfig


logger = logging.getLogger(__name__)


try:
    provisional_toolkit = ETSConfig.provisional_toolkit
except AttributeError:
    from contextlib import contextmanager

    # for backward compatibility
    @contextmanager
    def provisional_toolkit(toolkit_name):
        """ Perform an operation with toolkit provisionally set

        This sets the toolkit attribute of the ETSConfig object set to the
        provided value. If the operation fails with an exception, the toolkit
        is reset to nothing.
        """
        if ETSConfig.toolkit:
            raise AttributeError("ETSConfig toolkit is already set")
        ETSConfig.toolkit = toolkit_name
        try:
            yield
        except:
            # reset the toolkit state
            ETSConfig._toolkit = ''
            raise


def _init_toolkit():
    """ Initialise a toolkit, if possible. """

    def import_toolkit(tk):
        plugins = list(pkg_resources.iter_entry_points('pyface.toolkits', tk))
        if len(plugins) == 0:
            msg = "no Pyface plugin found for toolkit '{}'"
            raise RuntimeError(msg.format(tk))
        elif len(plugins) > 1:
            msg = ("multiple Pyface plugins found for toolkit %r: %s")
            modules = ', '.join(plugin.module_name for plugin in plugins)
            logger.warning(msg, tk, modules)

        exception = None
        while plugins:
            plugin = plugins.pop(0)
            try:
                tk_object = plugin.load()
                return tk_object
            except ImportError as exception:
                logger.exception(exception)
                msg = "Could not load plugin %r from %r"
                logger.warning(msg, plugin.name, plugin.module_name)
        else:
            # no success
            if exception is not None:
                raise exception
            else:
                raise RuntimeError("No toolkit loaded or exception raised.")

    # Get the toolkit.
    if ETSConfig.toolkit:
        return import_toolkit(ETSConfig.toolkit)

    # Try known toolkits first.
    for tk in ('qt4', 'wx'):
        try:
            with provisional_toolkit(tk):
                return import_toolkit(tk)
        except RuntimeError as exc:
            exc_info = logger.getEffectiveLevel() <= logging.INFO
            level = logger.ERROR if exc_info else logging.INFO
            msg = "Could not import Pyface backend %r"
            logger.log(level, msg, tk, excinfo=exc_info)

    # Try all non-null plugins we can find until success.
    for plugin in pkg_resources.iter_entry_points('pyface.toolkits'):
        if plugin.name == 'null':
            continue
        try:
            with provisional_toolkit(plugin.name):
                return plugin.load()
        except ImportError as exc:
            logger.exception(exc)
            msg = "Could not load plugin %r from %r"
            logger.warning(msg, plugin.name, plugin.module_name)

    # Try to import the null toolkit properly
    try:
        with provisional_toolkit('null'):
            return import_toolkit('null')
    except RuntimeError as exc:
        if logger.getEffectiveLevel() <= logging.INFO:
            logger.exception(exc)
        logger.info("Could not import Pyface backend 'null'")
    except ImportError:
        logger.info("Could not import Pyface backend 'null'")

    try:
        from pyface.ui.null.init import toolkit_object
        return toolkit_object
    except ImportError as exc:
        logger.exception(exc)
        raise ImportError("Unable to import a pyface backend for any toolkit")

    # if everything else fails toolkit is None
    return None


# The toolkit object function.
toolkit_object = _init_toolkit()





# Enthought library imports.
from envisage.ui.tasks.api import TaskWindow
from pyface.action.api import Action, ActionItem, Group
from pyface.tasks.api import Task
from pyface.tasks.action.api import TaskActionManagerBuilder, TaskActionController
from traits.api import Instance, Callable, Event, Str, List, Any, Property, cached_property, on_trait_change, Undefined

from .editor import FrameworkEditor

import logging
log = logging.getLogger(__name__)


BENCHMARK_OLD = False
if not BENCHMARK_OLD:
    # These are new implementations of actions that use far fewer traits than
    # the default enthought ones. The drawback is, of course, they aren't
    # updated the instant their dynamic trait is updated. The menu_update_event
    # on the Task object should be set during idle handling to update the menu.

    log.debug("BENCHMARKING THE FRAMEWORK ACTIONS!")
    class BasicFrameworkAction(Action):

        ##### Class attributes

        class DummyTask(Task):
            menu_update_event = Event
            active_editor = None

        dummy_task = DummyTask()

        ##### Traits

        # The Task with which the action is associated. Set by the framework.
        task = Instance(Task)

        @property
        def object(self):
            return None

        @property
        def active_editor(self):
            if self.task is not None:
                return self.task.active_editor
            return None

        ##### Protected interface.

        def _get_attr(self, obj, name, default=None):
            try:
                for attr in name.split('.'):
                    # Perform the access in the Trait name style: if the object
                    # is None, assume it simply hasn't been initialized and
                    # don't show the warning.
                    if obj is None:
                        return default
                    else:
                        obj = getattr(obj, attr)
            except AttributeError:
                if obj is Undefined or obj is None:
                    pass
                else:
                    log.error("Did not find name %r on %r" % (attr, obj))
                return default
            return obj

        ##### Action interface

        def destroy(self):
            # Disconnect listeners to task and dependent properties. Use the
            # dummy task instead of None so the intermediate trait handler
            # task.menu_update_event on on_dynamic_menu_update below won't
            # barf. Since it depends on self.task, it gets notified when
            # self.task changes, and if it gets set to None it gets called,
            # resulting in the error because it can't find the
            # menu_update_event on None.
            self.task = self.dummy_task

    class EnabledFrameworkAction(BasicFrameworkAction):
        # trait on the task that is used to set the enabled/disabled state.
        # Can't change dynamically; for efficiency purposes, the enabled name
        # as set at creation time will be used for the life of the action.
        #
        # Popup enable/checked state can be handled differently by defining the
        # methods _update_popup_enabled and _update_popup_checked. Those
        # methods take a single argument containing the popup data (the same
        # popup data as supplied to popup_context_menu_from_actions). The
        # enabled and checked state falls back to the _update_enabled and
        # _update_checked if the popup-specific methods are not supplied.
        enabled_name = Str("")

        @on_trait_change('task.menu_update_event')
        def on_dynamic_menu_update(self, ui_state):
            if ui_state is Undefined:
                return
            if self.active_editor:
                self._update_enabled(ui_state)
                self._update_checked(ui_state)
            self.on_dynamic_menu_update_hook(ui_state)
            log.debug("on_dynamic_menu_update %s: %s" % (self, self.enabled))

        def on_dynamic_menu_update_hook(self, evt):
            pass

        def on_popup_menu_update(self, ui_state, popup_data):
            try:
                self._update_popup_enabled(ui_state, popup_data)
            except AttributeError:
                self._update_enabled(ui_state)
            try:
                self._update_popup_checked(ui_state, popup_data)
            except AttributeError:
                self._update_checked(ui_state)
            log.debug("on_popup_menu_update %s: %s" % (self, self.enabled))

        def _update_enabled(self, ui_state):
            if self.enabled_name and ui_state:
                enabled = bool(self._get_attr(ui_state, self.enabled_name, None))
                if enabled is None:
                    log.warning("%s flag does not exist in ui_state object %s" % (self.enabled_name, ui_state))
                    enabled = False
                self.enabled = enabled
            else:
                self.enabled = bool(self.task)

        def _update_checked(self, ui_state):
            pass


    class NameChangeAction(EnabledFrameworkAction):
        menu_item_name = Str

        default_name = Str("-no event-")

        def on_dynamic_menu_update_hook(self, evt):
            if self.menu_item_name:
                if self.active_editor:
                    self.name = str(self._get_attr(self.active_editor,
                                                   self.menu_item_name, '0'))
                else:
                    self.name = getattr(self, 'default_name', '1')
            else:
                self.name = getattr(self, 'default_name', '2')
            log.debug("on_dynamic_menu_update_hook %s: %s" % (self.name, self.menu_item_name))

    EditorAction = EnabledFrameworkAction

if BENCHMARK_OLD:
    log.warning("BENCHMARKING THE ENTHOUGHT EDITOR ACTIONS!")
    from pyface.tasks.action.api import EditorAction
    FrameworkAction = EditorAction

    class IndividualTraitEventNonWorkingFrameworkAction(Action):
        # NOTE: this doesn't work, but I'm saving it in case I want to try
        # again with this method.


        # The Task with which the action is associated. Set by the framework.
        task = Instance(Task)

        # The active editor in the central pane with which the action is associated.
        active_editor = Property(Instance(FrameworkEditor),
                             depends_on='task.active_editor')

        # trait on the task that is used to set the enabled/disabled state.
        # Can't change dynamically; for efficiency purposes, the enabled name
        # as set at creation time will be used for the life of the action.
        enabled_name = Str("")

        ##### Protected interface.

        @cached_property
        def _get_active_editor(self):
            if self.task is not None:
                return self.task.active_editor
            return None

        def _get_attr(self, obj, name, default=None):
            try:
                for attr in name.split('.'):
                    # Perform the access in the Trait name style: if the object
                    # is None, assume it simply hasn't been initialized and
                    # don't show the warning.
                    if obj is None:
                        return default
                    else:
                        obj = getattr(obj, attr)
            except AttributeError:
                log.error("Did not find name %r on %r" % (attr, obj))
                return default
            return obj

        ##### Trait change handlers

        def _task_changed(self, old, new):
            method = getattr(self, '_enabled_update')
            name = getattr(self, 'enabled_name')
            print "ACTIVE", name, self.active_editor
            if name:
                if hasattr(self.active_editor, name):
                    name = "active_editor." + name
                    if old:
                        old.on_trait_change(method, name, remove=True)
                    if new:
                        new.on_trait_change(method, name)
                    print "ADDED ENABLED_NAME %s TO %s" % (name, self.active_editor)
            method()

        def _enabled_update(self):
            if self.enabled_name:
                if self.task:
                    self.enabled = bool(self._get_attr(self.active_editor, self.enabled_name, False))
                else:
                    self.enabled = False
            else:
                self.enabled = bool(self.task)
            print "ENABLING %s: %s" % (self.name, self.enabled)

        @property
        def object(self):
            return None

        # @property
        # def active_editor(self):
        #     if self.task is not None:
        #         return self.task.active_editor
        #     return None

        def destroy(self):
            # Disconnect listeners to task and dependent properties.
            self.task = None


    class EditorAction(EnabledFrameworkAction):
        pass


    class NameChangeAction(EditorAction):
        """Extension to the EditorAction that provides a user-updatable menu item
        name based on a trait
        
        EditorAction is subclassed from ListeningAction, and the ListeningAction
        methods destroy and _object_chaged must be called because the new
        trait name 'menu_item_name' can't be added to list of traits managed by
        ListeningAction and so must be taken care of here before the superclass
        can do its work.
        """
        menu_item_name = Str

        def destroy(self):
            """ Called when the action is no longer required.

            Remove all the task listeners.

            """

            if self.object:
                self.object.on_trait_change(
                    self._menu_item_update, self.menu_item_name, remove=True
                )
            super(NameChangeAction, self).destroy()

        def _menu_item_name_changed(self, old, new):
            obj = self.object
            if obj is not None:
                if old:
                    obj.on_trait_change(self._menu_item_update, old, remove=True)
                if new:
                    obj.on_trait_change(self._menu_item_update, new)
            self._label_update()

        def _object_changed(self, old, new):
            kind = 'menu_item'
            method = getattr(self, '_%s_update' % kind)
            name = getattr(self, '%s_name' % kind)
            if name:
                if old:
                    old.on_trait_change(method, name, remove=True)
                if new:
                    new.on_trait_change(method, name)
            method()
            super(NameChangeAction, self)._object_changed(old, new)

        def _menu_item_update(self):
            if self.menu_item_name:
                if self.object:
                    self.name = str(self._get_attr(self.object,
                                                   self.menu_item_name, 'Undo'))
                else:
                    self.name = getattr(self, 'default_name', 'Undo')
            else:
                self.name = getattr(self, 'default_name', 'Undo')


class BaseDynamicSubmenuGroup(Group):
    """ A group used for a dynamic menu.
    """

    #### 'ActionManager' interface ############################################

    id = 'DynamicMenuGroup'
    items = List

    #### 'TaskChangeMenuManager' interface ####################################

    # The ActionManager to which the group belongs.
    manager = Any

    # ENTHOUGHT QUIRK: This doesn't work: can't have a property depending on
    # a task because this forces task_default to be called very early in the
    # initialization process, before the window hierarchy is defined.
    #
    # active_editor = Property(Instance(IEditor),
    #                         depends_on='task.active_editor')

    event_name = Str('change this to the Event trait')

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_items(self, event_data=None):
        # Override this in your subclass to return the list of actions
        return []

    def _rebuild(self, new_trait_val):
        # Clear out the old group, then build the new one.
        self.destroy()

        # Get the new items, passing the event arguments to the method
        self.items = self._get_items(new_trait_val)

        # Set up parent so that radio items can determine their siblings to
        # uncheck others when checked. (see the _checked_changed method in
        # pyface/ui/wx/action/action_item.py)
        for item in self.items:
            item.parent = self

        # Inform our manager that it needs to be rebuilt.
        self.manager.changed = True

    #### Trait initializers ###################################################

    def _items_default(self):
        log.debug("DYNAMICGROUP: _items_default!!!")
        t = self._get_trait_for_event()
        t.on_trait_change(self._rebuild, self.event_name)
        return self._get_items()

    def _get_trait_for_event(self):
        raise NotImplementedError

    def _manager_default(self):
        manager = self
        while isinstance(manager, Group):
            manager = manager.parent
        log.debug("DYNAMICGROUP: _manager_default=%s!!!" % manager)
        return manager

    # ENTHOUGHT QUIRK: This doesn't work: the trait change decorator never
    # seems to get called, however specifying the on_trait_change in the
    # _items_default method works.
    #
    #    @on_trait_change('task.layer_selection_changed')
    #    def updated_fired(self, event):
    #        log.debug("SAVELAYERGROUP: updated!!!")
    #        self._rebuild(event)


class TaskDynamicSubmenuGroup(BaseDynamicSubmenuGroup):
    """ A group used for a dynamic menu.
    """

    # The task instance must be passed in as an attribute creation argument
    # because we need to bind on a task trait change to update the menu
    task = Instance('omnivore.framework.task.FrameworkTask')

    ###########################################################################
    # Private interface.
    ###########################################################################

    def _get_trait_for_event(self):
        return self.task

    #### Trait initializers ###################################################

    def _task_default(self):
        log.debug("DYNAMICGROUP: _task_default=%s!!!" % self.manager.controller.task)
        return self.manager.controller.task


class ApplicationDynamicSubmenuGroup(BaseDynamicSubmenuGroup):
    """ A group used for a dynamic menu based on an application event.
    """

    # The application instance must be a trait so we can set an on_trait_change
    # handler
    application = Instance('envisage.ui.tasks.api.TasksApplication')

    def _get_trait_for_event(self):
        return self.application

    #### Trait initializers ###################################################

    def _application_default(self):
        log.debug("DYNAMICGROUP: _application_default=%s!!!" % self.manager.controller.task.window.application)
        return self.manager.controller.task.window.application


class FrameworkTaskActionController(TaskActionController):
    # The whole point of this class is to redefine these two methods so that
    # the framework's custom action can have a task assigned to it. The parent
    # TaskActionController specifically checks for a TaskAction, which is the
    # problem. TaskActions subclass from (a few parents up) ListeningAction,
    # which is where I think the problem is where hundreds of thousands of
    # trait listener callbacks occur.
    def add_to_menu(self, item):
        """ Called when an action item is added to a menu/menubar.
        """
        action = item.item.action
        if hasattr(action, 'task'):
            action.task = self.task

    def add_to_toolbar(self, item):
        """ Called when an action item is added to a toolbar.
        """
        action = item.item.action
        if hasattr(action, 'task'):
            action.task = self.task


class FrameworkTaskActionManagerBuilder(TaskActionManagerBuilder):
    # Intercept the task action controller factory
    def _controller_default(self):
        return FrameworkTaskActionController(task=self.task)


class FrameworkTaskWindow(TaskWindow):
    # The only purpose of this is to override the envisage TaskWindow to
    # intercept the action manager builder factory so the framework can provide
    # its own.
    action_manager_builder_factory = Callable(FrameworkTaskActionManagerBuilder)



#------------------------------------------------------------------------------
#
#  Copyright (c) 2005, Enthought, Inc.
#  All rights reserved.
#
#  This software is provided without warranty under the terms of the BSD
#  license included in enthought/LICENSE.txt and may be redistributed only
#  under the conditions described in the aforementioned license.  The license
#  is also available online at http://www.enthought.com/licenses/BSD.txt
#
#  Thanks for using Enthought open source!
#
#  Author: David C. Morrill
#  Date:   10/14/2004
#
#------------------------------------------------------------------------------

""" Defines common traits used within the traits.ui package.
"""

#-------------------------------------------------------------------------
#  Imports:
#-------------------------------------------------------------------------

from __future__ import absolute_import

from traits.api import (Any, Delegate, Enum, Expression, Float,
                        HasStrictTraits, Instance, List, Range, Str, Trait,
                        TraitError, TraitPrefixList, TraitType)

from traits.trait_base import get_resource_path

#-------------------------------------------------------------------------
#  Trait definitions:
#-------------------------------------------------------------------------

# Orientation trait:
Orientation = Trait('vertical',
                    TraitPrefixList('vertical', 'horizontal'))

# Styles for user interface elements:
EditorStyle = style_trait = Trait(
    'simple',
    TraitPrefixList(
        'simple',
        'custom',
        'text',
        'readonly'),
    cols=4)

# Group layout trait:
Layout = Trait('normal',
               TraitPrefixList('normal', 'split', 'tabbed', 'flow', 'fold'))

# Trait for the default object being edited:
AnObject = Expression('object')

# The default dock style to use:
DockStyle = dock_style_trait = Enum('fixed', 'horizontal', 'vertical', 'tab',
                                    desc="the default docking style to use")

# The category of elements dragged out of the view:
ExportType = Str(desc='the category of elements dragged out of the view')

# Delegate a trait value to the object's **container** trait:
ContainerDelegate = container_delegate = Delegate('container',
                                                  listenable=False)

# An identifier for the external help context:
HelpId = help_id_trait = Str(desc="the external help context identifier")

# A button to add to a view:
AButton = Any
#AButton = Trait( '', Str, Instance( 'traitsui.menu.Action' ) )

# The set of buttons to add to the view:
Buttons = List(AButton,
               desc='the action buttons to add to the bottom of the view')

# View trait specified by name or instance:
AView = Any
#AView = Trait( '', Str, Instance( 'traitsui.view.View' ) )

# FIXME: on AButton and AView: TraitCompound handlers with deferred-import
# Instance traits are just broken. The Instance trait tries to update the
# top-level CTrait's fast_validate table when the import is resolved. However,
# sometimes the CTrait gets copied for unknown reasons and the copy's
# fast_validate table is not updated although the TraitCompound's slow_validates
# table is modified.

#-------------------------------------------------------------------------
#  'StatusItem' class:
#-------------------------------------------------------------------------


class StatusItem(HasStrictTraits):

    # The name of the trait the status information will be synched with:
    name = Str('status')

    # The width of the status field. The possible values are:
    #
    #   - abs( width )  > 1.0: Width of the field in pixels = abs( width )
    #   - abs( width ) <= 1.0: Relative width of the field when compared to
    #                          the other relative width fields.
    width = Float(0.5)

    def __init__(self, value=None, **traits):
        """ Initializes the item object.
        """
        super(StatusItem, self).__init__(**traits)

        if value is not None:
            self.name = value

#-------------------------------------------------------------------------
#  'ViewStatus' trait:
#-------------------------------------------------------------------------


class ViewStatus(TraitType):
    """ Defines a trait whose value must be a single StatusItem instance or a
        list of StatusItem instances.
    """

    # Define the default value for the trait:
    default_value = None

    # A description of the type of value this trait accepts:
    info_text = ('None, a string, a single StatusItem instance, or a list or '
                 'tuple of strings and/or StatusItem instances')

    def validate(self, object, name, value):
        """ Validates that a specified value is valid for this trait.
        """
        if isinstance(value, basestring):
            return [StatusItem(name=value)]

        if isinstance(value, StatusItem):
            return [value]

        if value is None:
            return value

        result = []
        if isinstance(value, SequenceTypes):
            for item in value:
                if isinstance(item, basestring):
                    result.append(StatusItem(name=item))
                elif isinstance(item, StatusItem):
                    result.append(item)
                else:
                    break
            else:
                return result

        self.error(object, name, value)


#-------------------------------------------------------------------------
#  'Image' trait:
#-------------------------------------------------------------------------------

image_resource_cache = {}
image_bitmap_cache = {}


def convert_image(value, level=3):
    """ Converts a specified value to an ImageResource if possible.
    """
    global image_resource_cache
    if not isinstance(value, str):
        return value

    key = value
    is_traits_image = (value[:1] == '@')
    if not is_traits_image:
        search_path = get_resource_path(level)
        key = '%s[%s]' % (value, search_path)
    result = image_resource_cache.get(key)

    if result is None:
        if is_traits_image:
            try:
                from pyface.image.image import ImageLibrary
                result = ImageLibrary.image_resource(value)
            except:
                result = None
        else:
            from pyface.image_resource import ImageResource
            result = ImageResource(value, search_path=[search_path])

        image_resource_cache[key] = result

    return result


def convert_bitmap(image_resource):
    """ Converts an ImageResource to a bitmap using a cache.
    """
    global image_bitmap_cache

    bitmap = image_bitmap_cache.get(image_resource)
    if (bitmap is None) and (image_resource is not None):
        #try:
        image_bitmap_cache[ image_resource ] = bitmap = \
            image_resource.create_bitmap()
        #except:
        #    pass

    return bitmap


class Image(TraitType):
    """ Defines a trait whose value must be a PyFace ImageResource or a string
        that can be converted to one.
    """

    # Define the default value for the trait:
    default_value = None

    # A description of the type of value this trait accepts:
    info_text = 'an ImageResource or string that can be used to define one'

    def __init__(self, value=None, **metadata):
        """ Creates an Image trait.

        Parameters
        ----------
        value : string or ImageResource
            The default value for the Image, either an ImageResource object,
            or a string from which an ImageResource object can be derived.
        """
        super(Image, self).__init__(convert_image(value), **metadata)

    def validate(self, object, name, value):
        """ Validates that a specified value is valid for this trait.
        """
        from pyface.image_resource import ImageResource

        if value is None:
            return None

        new_value = convert_image(value, 4)
        if isinstance(new_value, ImageResource):
            return new_value

        self.error(object, name, value)

    def create_editor(self):
        """ Returns the default UI editor for the trait.
        """
        from .editors.api import ImageEditor
        return ImageEditor()
#  'ATheme' trait:
#-------------------------------------------------------------------------


def convert_theme(value, level=3):
    """ Converts a specified value to a Theme if possible.
    """
    if not isinstance(value, basestring):
        return value

    if (value[:1] == '@') and (value.find(':') >= 2):
        try:
            from .image.image import ImageLibrary

            info = ImageLibrary.image_info(value)
        except:
            info = None

        if info is not None:
            return info.theme

    from .theme import Theme
    return Theme(image=convert_image(value, level + 1))


class ATheme(TraitType):
    """ Defines a trait whose value must be a traits UI Theme or a string that
        can be converted to one.
    """

    # Define the default value for the trait:
    default_value = None

    # A description of the type of value this trait accepts:
    info_text = 'a Theme or string that can be used to define one'

    def __init__(self, value=None, **metadata):
        """ Creates an ATheme trait.

        Parameters
        ----------
        value : string or Theme
            The default value for the ATheme, either a Theme object, or a
            string from which a Theme object can be derived.
        """
        super(ATheme, self).__init__(convert_theme(value), **metadata)

    def validate(self, object, name, value):
        """ Validates that a specified value is valid for this trait.
        """
        from .theme import Theme

        if value is None:
            return None

        new_value = convert_theme(value, 4)
        if isinstance(new_value, Theme):
            return new_value

        self.error(object, name, value)

#-------------------------------------------------------------------------------
#  'BasePMB' class:
#-------------------------------------------------------------------------------


class BaseMB(HasStrictTraits):
    def __init__(self, *args, **traits):
        """ Initializes the object.
        """
        n = len(args)
        if n > 0:
            if n == 1:
                left = right = top = bottom = args[0]
            elif n == 2:
                left = right = args[0]
                top = bottom = args[1]
            elif n == 4:
                left, right, top, bottom = args
            else:
                raise TraitError('0, 1, 2 or 4 arguments expected, but %d '
                                 'specified' % n)
            self.set(left=left, right=right, top=top, bottom=bottom)

        super(BaseMB, self).__init__(**traits)


#-------------------------------------------------------------------------------
#  'Margin' class:
#-------------------------------------------------------------------------------


class Margin(BaseMB):

    # The amount of padding/margin at the top:
    top = Range(-32, 32, 0)

    # The amount of padding/margin at the bottom:
    bottom = Range(-32, 32, 0)

    # The amount of padding/margin on the left:
    left = Range(-32, 32, 0)

    # The amount of padding/margin on the right:
    right = Range(-32, 32, 0)


#-------------------------------------------------------------------------------
#  'Border' class:
#-------------------------------------------------------------------------------


class Border(BaseMB):

    # The amount of border at the top:
    top = Range(0, 32, 0)

    # The amount of border at the bottom:
    bottom = Range(0, 32, 0)

    # The amount of border on the left:
    left = Range(0, 32, 0)

    # The amount of border on the right:
    right = Range(0, 32, 0)


#-------------------------------------------------------------------------------
#  'HasMargin' trait:
#-------------------------------------------------------------------------------


class HasMargin(TraitType):
    """ Defines a trait whose value must be a Margin object or an integer or
        tuple value that can be converted to one.
    """

    # The desired value class:
    klass = Margin

    # Define the default value for the trait:
    default_value = Margin(0)

    # A description of the type of value this trait accepts:
    info_text = (
        'a Margin instance, or an integer in the range from -32 to 32 '
        'or a tuple with 1, 2 or 4 integers in that range that can be '
        'used to define one')

    def validate(self, object, name, value):
        """ Validates that a specified value is valid for this trait.
        """
        if isinstance(value, int):
            try:
                value = self.klass(value)
            except:
                self.error(object, name, value)
        elif isinstance(value, tuple):
            try:
                value = self.klass(*value)
            except:
                self.error(object, name, value)

        if isinstance(value, self.klass):
            return value

        self.error(object, name, value)

    def get_default_value(self):
        """ Returns a tuple of the form: ( default_value_type, default_value )
            which describes the default value for this trait.
        """
        dv = self.default_value
        dvt = self.default_value_type
        if dvt < 0:
            if isinstance(dv, int):
                dv = self.klass(dv)
            elif isinstance(dv, tuple):
                dv = self.klass(*dv)

            if not isinstance(dv, self.klass):
                return super(HasMargin, self).get_default_value()

            self.default_value_type = dvt = 7
            dv = (self.klass, (), dv.get())

        return (dvt, dv)


#-------------------------------------------------------------------------------
#  'HasBorder' trait:
#-------------------------------------------------------------------------------


class HasBorder(HasMargin):
    """ Defines a trait whose value must be a Border object or an integer
        or tuple value that can be converted to one.
    """

    # The desired value class:
    klass = Border

    # Define the default value for the trait:
    default_value = Border(0)

    # A description of the type of value this trait accepts:
    info_text = (
        'a Border instance, or an integer in the range from 0 to 32 '
        'or a tuple with 1, 2 or 4 integers in that range that can be '
        'used to define one')
#  Other trait definitions:
#-------------------------------------------------------------------------
# The position of an image relative to its associated text:
Position = Enum('left', 'right', 'above', 'below')

# The alignment of text within a control:
Alignment = Enum('default', 'left', 'center', 'right')

# The spacing between two items:
Spacing = Range(-32, 32, 3)

#-------------------------------------------------------------------------
#  Other definitions:
#-------------------------------------------------------------------------

# Types that represent sequences:
SequenceTypes = (tuple, list)
